'''
*******************************************
LISA: Landscape In-Silico deletion Analysis
*******************************************

LISA is a statistical test that finds the most influential Transcription Factors related to a set of genes. We leverage integrative modeling of a comprehensive dataset 
of 100s of chromatin accessiblity samples and 1000s of ChIP experiments to make predictions. Particularly, LISA models how much the *cis*-regulatory elements around 
a gene are influenced by deleting elements associated with a TF (a process we call *insilico* deletion). For more information, see 
`<https://genomebiology.biomedcentral.com/articles/10.1186/s13059-020-1934-6>`_.

.. contents:: Interfaces

'''

from .lisa_user_data.regions_test import FromRegions
from .lisa_public_data.genes_test import FromGenes
from .lisa_user_data.coverage_test import FromCoverage
from .core.io import parse_regions_file, parse_macs_file

__version__ = '2.3.1'



from lisa.core.lisa_core import LISA_Core
from lisa.core.data_interface import PACKAGE_PATH, REQURED_DATASET_VERSION
from lisa.core.lisa_core import CONFIG_PATH as base_config_path
import numpy as np
from scipy import sparse
from lisa.core import genome_tools
import os
import configparser
from collections.abc import Iterable
from lisa.lisa_user_data.assays import ISD_Assay
from lisa.core.data_interface import DataInterface
from collections import Counter
from lisa.core.io import parse_macs_file, parse_regions_file

CONFIG_PATH = os.path.join(os.path.dirname(__file__),'config.ini')
_config = configparser.ConfigParser()
_config.read([base_config_path, CONFIG_PATH])

class ZeroScoreError(Exception):
    pass

class FromRegions(LISA_Core):
    '''
lisa.FromRegions
****************

Inputs:
    * Genes of interest
    * Regions (from bulk MACS2 peakcalling or peak-count matrix)
    * Region scores (optional)

Outputs:
    * Predicted TF influence

Interface for performing LISA test for TF influence using user-provided regions and genes-of-interest. The regions may be accompanied with a positive weight or score that
notes the strength of that region for your metric of interest. Often, that metric is ATAC-seq or DNase read depth at that region, but you may provide any 
score as long as it is positive. Regions should be more than 100 bp wide, but less than 1000 bp 
to ensure specificy for TF and motif hits within the regions. 

For optimal performance, your regions-of-interest should number > 20K and cover roughly the whole genome. If your regions are restricted to a certain chromosome,
You must manually provide background genes that are proximal to your regions.

This test also allows more flexibility to change LISA's function for mapping genomic regions' influence on nearby genes. By default, LISA uses 'Regulatory Potential' 
with a decay of 10000 bp, meaning the regions over a gene's TSS recieve maximum influence, and influence decays by half every 10K bp. This decay rate can be increased to 
allow long-range distal elements more weight, or reduced to prioritize promoter influence.

This interface outputs results in the same format as the ``FromGenes`` interface.

Example::

    with open('./genelist.txt', 'r') as genes_file:
        genes = [x.strip() for x in genes_file.readlines()]

    #Using Bedfile
    results, metadata = lisa.FromRegions.using_bedfile('hg38', genes, './path_to_bedfile.bed')

    #Using MACS output
    results, metadata = lisa.FromRegions.using_macs_output('hg38', './path_to_macs.xls', genes)

    #Using Estimator
    regions, scores = lisa.parse_bedfile('./path_to_bedfile.bed', header = False)
    results, metadata = lisa.FromRegions('hg38', regions).predict(genes)

    results_df = pd.DataFrame(results.to_dict())

For more, see `User Guide <user_guide.md>`_.

    '''

    @classmethod
    def get_docs(cls):
        return '\n'.join(x.__doc__ for x in 
        [cls, cls.using_bedfile, cls.using_macs_output, cls.__init__, cls.predict, cls.get_rp_map, cls.get_binding_matrix, parse_regions_file])

    @classmethod
    def using_macs_output(cls, species, query_genes, xls_path, rp_map = 'enhanced', rp_decay = 10000, isd_method = 'chipseq', 
            background_list = [], background_strategy = 'all', num_background_genes = 3000, seed = 2556, header = False, verbose = 4, log = None):
        '''
*classmethod*
**lisa.FromRegions.using_macs_output** (species, query_genes, xls_path, rp_map = 'enhanced', rp_decay = 10000, isd_method = 'chipseq', background_list = [], background_strategy = 'regulatory', num_background_genes = 3000, seed = 2556, header = False, verbose = 4, log = None)

    Use regions defined in MACS .xls file, and take the "pileup" field to be the region's score. 
    All arguments are the same as the "using_bedfile" method, except user must pass "xls_path" as path to MACS2 "{name}.xls" file.
    Header parameter has no effect.
        '''

        region_fields, region_scores = parse_macs_file(xls_path)
        
        lisa = cls(species, region_fields, rp_map = rp_map, rp_decay=rp_decay, isd_method=isd_method, verbose=verbose, log=log)

        return lisa.predict(query_genes, region_scores = region_scores, background_list=background_list, 
            background_strategy=background_strategy, num_background_genes=num_background_genes, seed=seed)

    @classmethod
    def using_bedfile(cls, species, query_genes, bed_path, rp_map = 'enhanced', rp_decay = 10000, isd_method = 'chipseq', 
            background_list = [], background_strategy = 'all', num_background_genes = 3000, seed = 2556, header = False, verbose = 4, log = None):
        '''
*classmethod*
**lisa.FromRegions.using_bedfile** (species, query_genes, bed_path, rp_map = 'enhanced', rp_decay = 10000, isd_method = 'chipseq', background_list = [], background_strategy = 'regulatory', num_background_genes = 3000, seed = 2556, header = False, verbose = 4, log = None)**

    Run LISA FromRegions test using a bedfile.

    Parameters:
        species: {'hg38', 'mm10'}

        query_genes (list): 
            Genes-of-interest, in either Symbol of RefSeqID format. Must provide between 20 to 500 genes.
        bed_path (str): 
            Path to tab-delineated bedfile with columns: chr start end [score]. The score column is optional.

    Returns:
        results (lisa.core.utils.LISA_Results): 
            With each key representing a table column, sorted by "summary_p_value" field. The dictionary can be passed directly to a the pandas constructor: ``results_df = pd.DataFrame(results.to_dict())``.
        metadata (dict): 
            Test metadata. Includes query genes provided and background genes that were selected, as well as reg-scores for top 100 factors on selected genes.
        '''
    
        assert(type(header) == bool)

        region_fields, region_scores = parse_regions_file(bed_path, header)

        lisa = cls(species, region_fields, rp_map = rp_map, rp_decay=rp_decay, isd_method=isd_method, verbose=verbose, log=log)

        return lisa.predict(query_genes, region_scores = region_scores, background_list=background_list, 
            background_strategy=background_strategy, num_background_genes=num_background_genes, seed=seed)


    window_size = 100

    def __init__(self, species, regions, rp_map = 'enhanced', rp_decay = 10000, isd_method = 'chipseq', verbose = 4, log = None):
        '''
*class*
**lisa.FromRegions** (species, regions, rp_map = 'enhanced', rp_decay = 10000, isd_method = 'chipseq', verbose = 4, log = None)**

    Initialize the LISA test using user-defined regions.

    Parameters:
        species: {'hg38', 'mm10'}

        regions (list of lists/tuples with format [('chr', start, end), ... ]):
            User-defined regions. 
        rp_map ({"basic", "enhanced"}, scipy.sparse_matrix):
            RP map type, currently supports "basic" and "enhanced". User may also pass their own RP map as scipy.sparse_matrix in the shape (genes x regions)
        rp_decay (float, int):
            Decay rate of region influence on gene based on distance from TSS. Increase to prioritize distal regions, decrease to prioritize promoters. Default of 10000 bp is balanced.
        isd_method {"chipseq", "motifs"}:
            Use ChIP-seq data or motifs to mark TF binding locations.
        verbose (int):
            Number of levels of log messages to print to stderr
    
    Returns:
        lisa object
        '''

        super().__init__(species, _config, 100, isd_method= isd_method, verbose=verbose, log = log)

        if isinstance(rp_map, str):
            rp_map_styles = self._config.get('lisa_params','rp_map_styles').split(',')
            assert(rp_map in rp_map_styles), 'RP map must be numpy/scipy.sparse array, or be one of provided maps: {}'.format(','.join(rp_map_styles))
        else:
            assert( isinstance(rp_map, np.ndarry) or isinstance(rp_map, scipy.sparse)), 'RP map must be either numpy ndarry or scipy.sparse matrix'
        self.rp_map = rp_map

        #self.genome = genome_tools.Genome.from_file(self._config.get('paths','genomes').format(package_path = PACKAGE_PATH, species = self.species), window_size=100)

        assert(isinstance(regions, (list, tuple))), '"regions" parameter must be list of region tuples in format [ (chr,start,end [,score]), (chr,start,end [,score]) ... ] or name of bed file.'
        
        self.log.append('Validating user-provided regions ...')
        self.num_regions_supplied = len(regions)

        regions = self._check_region_specification(regions)

        self.region_set = genome_tools.RegionSet(regions, self.data_interface.genome)
        self.region_score_map = np.array([r.annotation for r in self.region_set.regions])

        assert(isinstance(rp_decay, (int, float)) and rp_decay > 0), 'RP decay parameter must be positive int/float'
        self.rp_decay = rp_decay

        assert(len(regions) >= 1000 and len(regions) < 1000000), 'User must provide atleast 1000 reigons, and less than 1 million.'

    def _check_region_specification(self, regions):

        invalid_chroms = Counter()
        valid_regions = []
        for i, region in enumerate(regions):
            assert(isinstance(region, (tuple, list)) and len(region) == 3), 'Error at region #{}: Each region passed must be in format (string \"chr\",int start, int end'\
                .format(str(i))
            
            try:
                new_region = genome_tools.Region(*region, annotation = i)
                self.data_interface.genome.check_region(new_region)
                valid_regions.append(new_region)
            except ValueError:
                raise AssertionError('Error at region #{}: Could not coerce positions into integers'.format(str(i)))
            except genome_tools.NotInGenomeError as err:
                invalid_chroms[region[0]]+=1
                #raise AssertionError('Error at region #{}: '.format(str(i+1)) + str(err) + '\nOnly main chromosomes (chr[1-22,X,Y] for hg38, and chr[1-19,X,Y] for mm10) are permissible for LISA test.')
            except genome_tools.BadRegionError as err:
                raise AssertionError('Error at region #{}: '.format(str(i+1)) + str(err))
        
        if len(invalid_chroms) > 0 :
            self.log.append('WARNING: {} regions encounted from unknown chromsomes: {}'.format(
                str(sum(invalid_chroms.values())), str(','.join(invalid_chroms.keys()))
            ))

        return valid_regions
    
    def _load_factor_binding_data(self):

        self.factor_binding, self.factor_dataset_ids, self.factor_metadata = self.data_interface.get_binding_data(self.isd_method)

        m2m_region_map = np.array(self.region_set.map_genomic_windows(regions_to_bins=False)).astype(int)

        self.factor_binding = self.data_interface.project_sparse_matrix(self.factor_binding, m2m_region_map, 
            len(self.region_set), binarize=True)

        return self.factor_binding, self.factor_dataset_ids, self.factor_metadata

    def _load_rp_map(self):

        if isinstance(self.rp_map, str):
            if self.rp_map == 'basic':
                self.rp_map = self.data_interface._make_basic_rp_map(self.data_interface.gene_loc_set, self.region_set, self.rp_decay)
            elif self.rp_map == 'enhanced':
                self.rp_map = self.data_interface._make_enhanced_rp_map(self.data_interface.gene_loc_set, self.region_set, self.rp_decay)
            else:
                NotImplementedError()
        else:
            desired_shape = (len(self.data_interface.gene_loc_set), len(self.num_regions_supplied))
            assert(self.rp_map.shape == desired_shape), 'RP_map must be of shape (n_genes, n_regions): {}'.format(str(desired_shape))

        return self.rp_map

    def _initialize_assays(self, **assay_kwargs):

        self.add_assay(
            ISD_Assay(**assay_kwargs, technology = 'Regions')
        )

    def get_rp_map(self):
        '''
    *method*
    **.get_rp_map** ()

        Return RP map calculated for your regions, along with the gene and regions metadata.

        Returns:
            rp_map (scipy.sparse_matrix):
                The calculated RP map, (genes x regions)
            gene_metdata (list of tuples):
                List of genes, with columns chr,start,end,symbol
            region_metadata (list of tuples):
                List of supplied regions, sorted. Columns are chr,start,end

        Formatting to Anndata::

            lisa_test = lisa.FromRegions('hg38', regions)
            rp_map, gene_metadata, region_metadata = lisa_test.get_rp_map()

            rp_map_anndata = anndata.AnnData(X = rp_map, 
                obs = pd.DataFrame(gene_metadata, columns = ['chr','start','end','symbol']),
                var = pd.DataFrame(region_metadata, columns = ['chr','start','end'])
            )
        '''
        if isinstance(self.rp_map, str):
            self._load_rp_map()
        
        gene_loc_metadata = [(gene.chromosome, gene.start, gene.end, gene.annotation.get_name(), gene.annotation.strand) for gene in self.data_interface.gene_loc_set.regions]
        regions_metadata = [r.to_tuple() for r in self.region_set.regions]

        return self.rp_map, gene_loc_metadata, regions_metadata

    def get_binding_matrix(self):
        '''
    *method*
    **.get_binding_matrix** ()

        Returns a binary matrix of factor hits within your regions. Rows are regions, columns are either ChIP-seq samples or motifs. 

        Returns:
            factor_binding_matrix (scipy.sparse_matrix):
                Sparse binary matrix of shape (regions x factor). Is "1" if factor is predicted to bind at region, "0" if not.
            region_metadata (list of tuples):
                List of supplied regions, sorted. Columns are chr,start,end
            factor_metadata (dict of lists):
                Metadata for each factor binding profile.

        Formatting to Anndata::

            lisa_test = lisa.FromRegions('hg38', regions)
            factor_binding, regions, factors = lisa_test.get_binding_matrix()

            binding_anndata = anndata.AnnData(X = factor_binding,
                obs = pd.DataFrame(regions, columns = ['chr','start','end'])
                var = pd.DataFrame(factors)
            )

        '''

        try:
            self.factor_binding
        except AttributeError:
            self._load_factor_binding_data()

        regions_metadata = [r.to_tuple() for r in self.region_set.regions]

        return self.factor_binding, regions_metadata, self.factor_metadata


    def predict(self, query_genes, region_scores = None, background_list = [], background_strategy = 'all', num_background_genes = 3000, seed = 2556):
        '''
    *method*
    **.predict** (query_genes, region_scores = None, background_list = [], background_strategy = 'all', num_background_genes = 3000, seed = 2556)**
    
        Predict TF influence given a set of genes.
        
        Params:
            query_genes (list):
                Genes-of-interest, in either Symbol of RefSeqID format. Must provide between 20 to 500 genes.
            region_scores (list or np.ndarray of shape (len(regions), ):
                Region scores/weights. Must be same length as regions. If not passed, all regions will be given score of 1.
            background_list (list):
                User-specified list of background genes to compare with query_list. Must contain more genes than query list and entire list will be used. If provided, ```background_strategy``` must be set to "provided".
            background_strategy {"regulatory","random","provided"}:
                Regulatory will sample background genes from a stratified sample of TADs and regulatory states, random will randomly sample from all non-query genes.
            num_background_genes (int):
                Number of genes to use as comparison to query genes. More background genes make test slower, but more stable.
            seed (int):
                Seed for gene selection and regression model initialization.

        Returns
            results:
                lisa.core.utils.LISA_Results with each key representing a table column, sorted by "summary_p_value" field. The results can be passed directly to a the pandas constructor by calling the "to_dict()" command: ``results_df = pd.DataFrame(results.to_dict())``.
            metadata: 
                Test metadata. Includes query genes provided and background genes that were selected, as well as reg-scores for top 100 factors on selected genes.
        '''

        if region_scores is None:
            region_scores = np.ones(self.num_regions_supplied)

        else:
            assert(isinstance(region_scores, (np.ndarray, list))), 'Passed scores must be of type list or numpy.ndarray'
            assert(len(region_scores) == self.num_regions_supplied), 'Must provided a score for each region passed. Proved {} regions, and {} scores'\
                .format(str(len(self.region_set)), str(self.num_regions_supplied))
            
            try:
                region_scores = np.array(region_scores).astype(np.float64)
            except ValueError as err:
                raise AssertionError('Region score could not be cast to float: ' + str(err))

            assert(len(region_scores.shape) == 1), 'Region scores must be 1-D list or array.'
            
            assert(np.all(region_scores >= 0)), 'All scores must be non-negative'

            region_scores = region_scores[self.region_score_map]

        return super().predict(query_genes, background_list=background_list, background_strategy=background_strategy, 
            num_background_genes= num_background_genes, seed=seed, region_scores = region_scores)


from lisa.core.utils import LoadingBar
from lisa.core.assays import LISA_RP_Assay, get_delta_RP, get_deltaRP_activation
from multiprocessing import Pool
import numpy as np
from scipy import stats, sparse
import os

class ISD_Assay(LISA_RP_Assay):

    def predict(self, gene_mask, label_vector,*,region_scores,**kwargs):

        with self.log.section('Modeling insilico deletions:'):

            profile = (region_scores/region_scores.sum() * 1e5)[:,np.newaxis]

            subset_profile, subset_factor_binding, subset_rp_map = \
                self.make_subset_rp_map(rp_map = self.rp_map, factor_binding = self.factor_binding,
                    gene_mask = gene_mask, accessibility = profile)

            self.log.append('Calcuating null RP model ...')

            null_rp_matrix = self.make_rp_matrix(profiles = subset_profile, rp_map = subset_rp_map)

            self.log.append('Performing knockouts ...')
            delta_RP = get_delta_RP(subset_profile, subset_factor_binding, subset_rp_map)

            self.log.append('Calculating Δ regulatory score ...')
            delta_reg_scores = get_deltaRP_activation(null_rp_matrix, delta_RP)

            self.log.append('Calculating p-values ...')
            p_vals = self.get_delta_RP_p_value(delta_reg_scores, label_vector)

            query_reg_score_matrix, background_reg_score_matrix, top_factor_idx = self.get_delta_reg_score_matrix(p_vals, delta_reg_scores, label_vector)

            '''self.debug = dict(
                profile = profile,
                gene_mask = gene_mask,
                label_vector = label_vector,
                subset_rp_map = subset_rp_map,
                null_rp_matrix = null_rp_matrix,
                delta_RP_matrix = delta_RP,
                delta_reg_scores = delta_reg_scores,
            )'''

        self.log.append('Done!')

        return p_vals, dict(
            dataset_ids = list(np.array(self.factor_dataset_ids)[top_factor_idx]),
            query_reg_scores = query_reg_score_matrix.tolist(),
            background_reg_scores = background_reg_score_matrix.tolist()
        )
        


from lisa.core.lisa_core import LISA_Core
from lisa.lisa_public_data.genes_test import FromGenes
from lisa.core.data_interface import PACKAGE_PATH, REQURED_DATASET_VERSION, INSTALL_PATH
from lisa.core.lisa_core import CONFIG_PATH as base_config_path
import numpy as np
from scipy import sparse
from lisa.core import genome_tools
import os
import configparser
from collections.abc import Iterable
from lisa.lisa_user_data.assays import ISD_Assay
from lisa.core.data_interface import DataInterface
from collections import Counter
from lisa.core.utils import Log, LoadingBar

CONFIG_PATH = os.path.join(os.path.dirname(__file__),'config.ini')
_config = configparser.ConfigParser()
_config.read([base_config_path, CONFIG_PATH])


class FromCoverage(LISA_Core):
    '''
lisa.FromCoverage
****************

Inputs:
    * Genes of interest
    * BigWig file, coverage over genome

Outputs:
    * Predicted TF influence

Use LISA to infer TF influence on your geneset using your own coverage data. This test is better suited than the "regions" test when your measure produces wide peaks/areas of influence.
An example of this is H3K27ac data, which correlates with gene expression similarly to accessibility, but produces wide peaks that may span many distinct TF binding locations.

This interface outputs results in the same format as the ``FromGenes`` interface.

Example::

    with open('./genelist.txt', 'r') as genes_file:
        genes = [x.strip() for x in genes_file.readlines()]

    results, metadata = lisa.FromRegions.using_bigwig('hg38', genes, './sample.bigwig')

    results_df = pd.DataFrame(results.to_dict())

For more, see `User Guide <user_guide.md>`_.
    '''

    window_size = FromGenes.window_size

    @classmethod
    def get_docs(cls):
        return '\n'.join(x.__doc__ for x in 
        [cls, cls.using_bigwig, cls.__init__, cls.predict])

    @classmethod 
    def using_bigwig(cls, species, query_genes, bigwig_path, rp_map = 'enhanced_10K', isd_method = 'chipseq', background_list = [], 
        background_strategy = 'all', num_background_genes = 3000, seed = 2556, verbose = 4, log = None):
        '''
*classmethod*
**lisa.FromCoverage.using_bigwig** (species, query_genes, bigwig_path, rp_map = 'basic', rp_decay = 10000, isd_method = 'chipseq', background_list = [], background_strategy = 'all', num_background_genes = 3000, seed = 2556, header = False, verbose = 4, log = None)

    Run LISA FromCoverage test using a bigwig coverage file.

    Parameters:
        species: {'hg38', 'mm10'}

        query_genes (list): 
            Genes-of-interest, in either Symbol of RefSeqID format. Must provide between 20 to 500 genes.
        bigwig_path (str): 
            Path to bigwig file

    Returns:
        results (lisa.core.utils.LISA_Results): 
            With each key representing a table column, sorted by "summary_p_value" field. The dictionary can be passed directly to a the pandas constructor: ``results_df = pd.DataFrame(results.to_dict())``.
        metadata (dict): 
            Test metadata. Includes query genes provided and background genes that were selected.
        '''

        if log is None:
            log = Log()
        
        coverage_array = cls.convert_bigwig(bigwig_path, species, log = log)

        return cls(species, coverage_array, rp_map = rp_map, isd_method=isd_method, verbose=verbose, log=log)\
            .predict(query_genes, background_list=background_list, background_strategy=background_strategy, num_background_genes=num_background_genes,
            seed=seed)

    @classmethod
    def convert_bigwig(cls, bigwig, species, log = None):

        try:
            import pyBigWig as bw
        except ImportError:
            raise ModuleNotFoundError('pyBigWig is required to use this command, but as of LISA 2.3.0 it is not'
                                      ' installed with the base package because its compiled C components'
                                      ' cause problems on some systems. To install the the version of LISA'
                                      ' where this command can be used, do:'
                                      '\n\n\t $ pip install lisa2[coverage]\n')

        if log is None:
            log = Log()

        genome = DataInterface.load_genome(species, cls.window_size)
        coverage_array = np.zeros(len(genome))

        log.append('Converting BigWig file to coverage array ...')

        bar = LoadingBar('Progress', len(genome) // 1000 + 1, cold_start=True)

        try:
            coverage_bw = bw.open(bigwig)

            log.append(bar, update_line=True)

            for i, window in enumerate(genome.list_windows()):
                
                if window.chromosome in coverage_bw.chroms():
                    mean_coverage = coverage_bw.stats(*window.to_tuple())[0]
                    coverage_array[i] = mean_coverage

                if i%1000 == 0:
                    log.append(bar, update_line = True)

            return np.nan_to_num(coverage_array)

        finally:
            coverage_bw.close()

    
    def __init__(self, species, coverage_array, rp_map = 'enhanced_10K', isd_method = 'chipseq', verbose = 4, log = None):
        '''
*class*
**lisa.FromCoverage** (species, regions, rp_map = 'enhanced_10K', rp_decay = 10000, isd_method = 'chipseq', verbose = 4, log = None)

    Initialize the LISA test using user-defined regions.

    Parameters:
        species: {'hg38', 'mm10'}

        coverage_array: (1D or Nx1 np.ndarray):
            Array of scores over 1kb bins.
        isd_method {"chipseq", "motifs"}:
            Use ChIP-seq data or motifs to mark TF binding locations.
        rp_map {"basic_10K", "enhanced_10K"}:
            Choice of RP map, which maps the regulatory influence of a region to a gene. The "basic_10K" model is based simply off distance, with the "enhanced_10K" model masks out the promoter and exon regions of other nearby genes.
        verbose (int):
            Number of levels of log messages to print to stderr
    
    Returns:
        lisa object
        '''

        super().__init__(species, _config, self.window_size, isd_method= isd_method, verbose=verbose, log = log)
        
        assert(isinstance(coverage_array, np.ndarray)), 'Coverage array must be of type numpy.ndarray'
        assert(len(coverage_array.shape) == 1 or (1 in coverage_array.shape and len(coverage_array.shape) == 2)), 'Coverage array must be 1D array or column/row vector'

        coverage_array = coverage_array.reshape(-1)
        assert(len(coverage_array) == len(self.data_interface.genome)), 'Coverage array must be of same length as genome: {} bins'.format(str(len(self.data_interface.genome)))
        self.coverage_array = coverage_array

        rp_map_styles = self._config.get('bam_test_params','rp_map_styles').split(',')
        assert(rp_map in rp_map_styles), 'RP map must be numpy/scipy.sparse array, or be one of provided maps: {}'.format(','.join(rp_map_styles))
        self.rp_map_style = rp_map

    def _load_factor_binding_data(self):
        return self.data_interface.get_binding_data(self.isd_method)
        
    def _load_rp_map(self):
        return self.data_interface.get_rp_map(self.rp_map_style)

    def _initialize_assays(self, **assay_kwargs):
        self.add_assay(
            ISD_Assay(**assay_kwargs, technology = 'BamTest')
        )

    def predict(self, query_list, background_list = [], background_strategy = 'all', num_background_genes = 3000, 
        seed = 2556):
        '''
    *method*
    **.predict** (self, query_list, background_list = [], background_strategy = 'all', num_background_genes = 3000, seed = 2556)
    
        Predict TF influence given a set of genes.
        
        Params:
            query_list (list):
                Genes-of-interest, in either Symbol of RefSeqID format. Must provide between 20 to 500 genes.
            background_list (list):
                User-specified list of background genes to compare with query_list. Must contain more genes than query list and entire list will be used. If provided, ```background_strategy``` must be set to "provided".
            background_strategy {"regulatory","random","provided"}:
                Regulatory will sample background genes from a stratified sample of TADs and regulatory states, random will randomly sample from all non-query genes.
            num_background_genes (int):
                Number of genes to use as comparison to query genes. More background genes make test slower, but more stable.
            seed (int):
                Seed for gene selection and regression model initialization.

        Returns:
            results (lisa.core.utils.LISA_Results):
                Can be passed directly to a the pandas constructor: ``results_df = pd.DataFrame(results.to_dict())``.
            metadata (dict):
                Test metadata. Includes query genes provided and background genes that were selected, as well as reg-scores for top 100 factors on selected genes.
        '''
        return super().predict(query_list, region_scores = self.coverage_array, background_list=background_list, background_strategy=background_strategy, 
            num_background_genes= num_background_genes, seed=seed)


from itertools import product
import subprocess
import os
import sys
import argparse

TEST_PATH = os.path.join('blah', 'test_files')

common_options = dict(
        use_motifs = ['','--use_motifs'],
        save_metadata = ['', '--save_metadata'],
        assays = ['--assays ' + a for a in ['DNase H3K27ac Direct', 'Direct', 'DNase']],
        verbose = '-v 3',
    )

new_test_badge = '''
***********************************
Test #{i}
{command}
***********************************
'''

def permute_options(options):
    all_permutations = []
    for permute_option, values in options.items():
        if isinstance(values, list) and len(values) > 1:
            permutations = values[1:]
            for option, values in options.items():
                if option != permute_option and values[0] != '':
                    if isinstance(values, list):
                        add_option = values[0]
                    else: 
                        add_option = values
                    permutations = [permutation + ' ' + add_option for permutation in permutations]
            all_permutations.extend(permutations)
    return all_permutations

def exec_commands(commands, run = 'all'):

    failed_commands = 0
    for i, command in enumerate(commands):
        print(new_test_badge.format(command = command, i = str(i + 1)))
        try:
            process = subprocess.run(command.split(' '), capture_output=True)
            if process.returncode!=0:
                raise AssertionError()
        except AssertionError:
            print(process.stderr.decode('utf-8'))
            failed_commands += 1
            pass
        else:
            print('Passed!')

    return failed_commands

def run_tests(base_command, options):
    
    cmd_combos = [base_command + ' ' + option for option in permute_options(options)]

    failed = exec_commands(cmd_combos)

    print('Passed {}/{} tests.'\
        .format(str(len(cmd_combos) - failed), str(len(cmd_combos))))

def test_oneshot(test_genelist, background_list):

    oneshot_options = dict(
        background_strategy = ['--background_strategy regulatory -b 501', 
            '--background_strategy random -b 300', 
            '--background_strategy provided --background_list {}'.format(background_list)],
        output_prefix = ['', '-o results'],
        **common_options
    )

    run_tests('lisa oneshot hg38 ' + test_genelist, oneshot_options)
    run_tests('lisa oneshot mm10 ' + test_genelist, oneshot_options)

def test_multi(genelists):

    mulit_options = dict(
        num_background_genes = '-b 501',
        use_random_background = ['', '--random_background'],
        output = '-o results/',
        **common_options
    )

    run_tests('lisa multi hg38 ' + ' '.join(genelists), mulit_options)
    run_tests('lisa multi mm10 ' + ' '.join(genelists), mulit_options)


import unittest
from lisa.core import data_interface
import numpy as np
import tempfile
import os
from scipy import sparse

class TestProjections(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.data = data_interface.DataInterface('mm10', window_size=1000, path = './test.h5', make_new=True)

    def test_array_projection(self):

        proj = np.array([
            (2,0),
            (3,1),
            (5,3),
        ])

        arr = np.array([0,0,1,2,0,3])
        correct = correct = np.array([1,2,0,3])
    
        self.assertTrue(
            np.all(self.data.project_array(arr, proj, 4) == correct)
        )

    def test_indices_projection(self):

        ind = [0,4,3]

        proj = np.array([
            (0,1),
            (1,2),
            (3,4),
            (4,5)
        ])

        correct = [1,5,4]

        ans = self.data.project_indices(ind, proj)

        self.assertTrue(
            len(set(ans)) == len(set(correct))
        )

        self.assertTrue(
            len(set(correct).intersection(set(ans))) == 3
        )

    def test_matrix_projection(self):

        arr = sparse.csr_matrix(np.array([
            (0,1),
            (1,0),
            (2,3),
            (3,4),
            (6,8),
            ])
        )

        proj = np.array([
            (0,1),
            (1,0),
            (4,2),
            (2,3),
        ])

        ans = np.array(self.data.project_sparse_matrix(arr, proj, 5).todense())

        correct = np.array([
            (1,0),
            (0,1),
            (6,8),
            (2,3),
            (0,0),
        ])

        self.assertTrue(
            np.all(correct == ans)
        )

    @classmethod
    def tearDownClass(cls):
        os.remove('./test.h5')


class TestDataConsistency(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.data = data_interface.DataInterface('mm10', window_size=100000, path = './test.h5', make_new=True)

    @staticmethod
    def sparse_matrix_equal(m1, m2):

        m1, m2 = m1.tocsr(), m2.tocsr()

        return np.all(m1.indices == m2.indices) and np.all(m1.indptr == m2.indptr) and \
            np.all(np.isclose(m1.data, m2.data))

    def test_rp_map_consistency(self):

        rp_map_shape = self.data.get_rp_map_shape()

        #random_rp_map = sparse.random(*rp_map_shape, density = 0.01)
        num_entries = 1000
        i = np.random.choice(rp_map_shape[0], num_entries, replace = False)
        j = np.random.choice(rp_map_shape[1], num_entries, replace = False)
        data = np.random.randn(num_entries)

        random_rp_map = sparse.csr_matrix((data, (i,j)), shape = rp_map_shape)

        self.data.add_rp_map('test', random_rp_map)

        self.assertTrue(
            tuple(self.data.get_rp_maps()) == ('test',)
        )

        reloaded = self.data.get_rp_map('test')

        self.assertTrue(
            self.sparse_matrix_equal(random_rp_map, reloaded)
        )

    def test_factor_binding_consistency(self):

        num_bins = len(self.data.genome)

        simulated_hits = np.random.choice(num_bins, 1000, replace = False)

        self.data.add_binding_data('ChIP-seq', '1', simulated_hits, factor = 'STAT', tissue = 'intestine',
            cell_line = '1893A', cell_type = 'blah', qc  = 1)

        reloaded, meta = self.data.get_binding_dataset('ChIP-seq', '1')

        self.assertTrue(
            len(set(simulated_hits).symmetric_difference(set(reloaded))) == 0
        )

        reloaded_matrix, ids, meta = self.data.get_binding_data('ChIP-seq')

        sim_hits_matrix = sparse.csc_matrix((np.ones_like(simulated_hits), simulated_hits, [0, len(simulated_hits)]),
                shape = (len(self.data.genome), 1))

        self.assertTrue(
            self.sparse_matrix_equal(sim_hits_matrix, reloaded_matrix)
        )

        self.assertTrue(
            tuple(ids) == ('1',)
        )

    def test_profile_consistency(self):

        simulated_profile = np.abs(np.random.randn(len(self.data.genome)))[:, np.newaxis]

        num_genes, num_bins = self.data.get_rp_map_shape()

        diag_rp_map = sparse.diags(
            np.ones(num_genes), shape = (num_genes, num_bins)
        )

        self.data.add_profile_data('DNase', '1', simulated_profile, [diag_rp_map], ['diag'], tissue = 'intestine',
            cell_line = '1893A', cell_type = 'blah', qc  = 1, norm_depth=None)

        reloaded, meta = self.data.get_profile('DNase', '1')

        self.assertTrue(
            np.all(np.isclose(simulated_profile, reloaded, rtol=1e-2))
        )

        rp_matrix, dataset_ids = self.data.get_rp_matrix('DNase','diag')

        self.assertTrue(
            np.all(np.isclose(rp_matrix[:num_bins], simulated_profile[:num_bins], rtol=1e-2))
        )


    @classmethod
    def tearDownClass(cls):
        os.remove('./test.h5')


if __name__ == '__main__':
    unittest.main()


import os
from lisa.core.lisa_core import LISA_Core
from lisa.core.lisa_core import CONFIG_PATH as base_config_path
from lisa.core.data_interface import PACKAGE_PATH, REQURED_DATASET_VERSION
import lisa.lisa_public_data.assays as assays
from lisa.lisa_public_data.models import LR_BinarySearch_SampleSelectionModel, LR_ChromatinModel
from lisa.core.data_interface import DataInterface
import numpy as np
from scipy import sparse
import multiprocessing
import configparser

CONFIG_PATH = os.path.join(os.path.dirname(__file__),'config.ini')
_config = configparser.ConfigParser()
_config.read([base_config_path, CONFIG_PATH])

class FromGenes(LISA_Core):
    '''
lisa.FromGenes
**************

Inputs:
    * Genes of interest

Outputs:
    * Predicted TF influence

Interface for performing LISA test for TF influence using public chromatin accessibility data. Given just a set of genes, LISA will identify a subset of a large database
of public H3K27ac and DNase profiles that represents a model for multiple chromatin states around those genes. LISA then assesses the influence of TF binding 
on your genes-of-interest vs a sampling of background genes through through those representative datasets, and aggregates the effects to produce a final p-value.

This model is useful for integrating accessibility and binding data when you have strictly a list of associated genes (from scRNA-seq, for example). If you have 
genes-of-interest, as well as regions-of-interest, you may use the more specific test provided by ``lisa.FromRegions``.

Example::

    with open('./genelist.txt', 'r') as genes_file:
        genes = [x.strip() for x in genes_file.readlines()]

    results, metadata = lisa.FromGenes('hg38').predict(genes)

    results_df = pd.DataFrame(results.to_dict())

For more, see `user guide <user_guide.md>`_.

    '''

    @classmethod
    def get_docs(cls):
        return '\n'.join(x.__doc__ for x in 
        [cls, cls.__init__, cls.predict])

    window_size = 1000
    
    def __init__(self, species, rp_map = 'enhanced_10K', assays = ['Direct','H3K27ac','DNase'], isd_method = 'chipseq', verbose = True, log = None):
        '''
*class*
**lisa.FromGenes** (species, rp_map = 'enhanced_10K', assays = ['Direct','H3K27ac','DNase'], isd_method = 'chipseq', verbose = True, log = None)

    Initialize the LISA test using public data.

    Params:
        species {'hg38', 'mm10'}:
        assays (list of {"Direct","H3K27ac","DNase"}):
            default is all tests
        isd_method {"chipseq", "motifs"}:
            Use ChIP-seq data or motifs to mark TF binding locations.
        rp_map {"basic_10K", "enhanced_10K"}:
            Choice of RP map, which maps the regulatory influence of a region to a gene. The "basic_10K" model is based simply off distance, with the "enhanced_10K" model masks out the promoter and exon regions of other nearby genes.
        verbose (int):
            Number of levels of log messages to print to stderr

    Returns:
        lisa object
        '''

        cores = 1
        super().__init__(species, _config, self.window_size, isd_method= isd_method, verbose=verbose, log=log)
        assert(len(assays) > 0), 'Must provide at least one assay to run.'
        assert(all([assay in self._config.get('lisa_params','assays').split(',') for assay in assays])), 'An assay chosen by the user is not a valid choice: \{{}}'.format(self._config.get('lisa_params','assays'))

        num_datasets_selected = int(_config.get('lisa_params', 'num_datasets_selected'))
        num_datasets_selected_anova = int(self._config.get('lisa_params','num_datasets_selected_anova'))

        assert( isinstance(num_datasets_selected, int) )
        assert( isinstance(num_datasets_selected_anova, int) )

        self.num_datasets_selected_anova = num_datasets_selected_anova
        self.num_datasets_selected = num_datasets_selected

        self.cores = self._set_cores(cores)

        assert( num_datasets_selected_anova > num_datasets_selected ), 'Anova must select more datasets than the regression model'
        assert( num_datasets_selected_anova > 0 and num_datasets_selected > 0 ), 'Number of datasets selected must be positive'
        assert(num_datasets_selected_anova < 500)
        assert(num_datasets_selected < 25)

        if self.num_datasets_selected % self.cores != 0:
            self.log.append('''WARNING: You\'ve allocated {} cores with {} datasets selected.
To ensure maximum speed, allocate as many cores as datasets selected.
For better efficiency, make #datasets a multiple of #cores.'''.format(self.cores, self.num_datasets_selected))

        self.schedule_assays = sorted(list(set(assays)))

        rp_map_styles = self._config.get('lisa_params','rp_map_styles').split(',')
        assert(rp_map in rp_map_styles), 'RP map must be numpy/scipy.sparse array, or be one of provided maps: {}'.format(','.join(rp_map_styles))
        self.rp_map_style = rp_map

    def _set_cores(self, cores):
        assert( isinstance(cores, int) and cores >= -1)
        #leave one core out for the rest of us
        max_cores = multiprocessing.cpu_count() - 1
        if cores <= -1:
            cores = max_cores
        #use the minimum number of cores between the user's specificaion, the number of datasets to be processed in parallel, and the number of cores on the machine.
        #this prevents LISA from using more resources than required.
        self.cores = min(cores, max_cores, self.num_datasets_selected)
        return self.cores

    def _load_factor_binding_data(self):
        return self.data_interface.get_binding_data(self.isd_method)
        
    def _load_rp_map(self):
        return self.data_interface.get_rp_map(self.rp_map_style)

    def _get_factor_gene_mask(self):

        factor_genes = self.data_interface.genes.match_user_provided_genes(self.factor_metadata['factor'])

        loc_symbol_dict = dict(zip(factor_genes.get_locations(), factor_genes.get_symbols()))
        factor_gene_mask = np.isin(self.data_interface.rp_map_locs, factor_genes.get_locations())
        #make label vector from label_dict based of gene_loc ordering in rp_map data
        factor_mask_keys = [loc_symbol_dict[gene_loc] for gene_loc in self.data_interface.rp_map_locs[self.factor_gene_mask]]

        return factor_gene_mask, factor_mask_keys

    '''def _format_results(self, assay_pvals, assay_info, gene_info_dict, **kwargs):
        
        new_columns = {}
        for assay, info in assay_info.items():
            if 'factor_acc_z_scores' in info:
                z_scores = info.pop('factor_acc_z_scores')
                z_dict = dict(zip(self.factor_mask_keys, z_scores))
                new_columns[assay + 'factor_accessibility_z_score'] = [z_dict.get(factor, 'NA') for factor in self.factor_metadata['factor']]

        return super()._format_results(assay_pvals, assay_info, gene_info_dict, **new_columns, **kwargs)'''

    #CHANGE THIS TO USE NEW DATA INTERFACE
    def _initialize_assays(self, **assay_kwargs):
        #Add assays to LISA's steps. Each assay follows the same instantiation and prediction calling, making them modular and substitutable.
        #Adding an assay loads the required data for that assay

        try: # if factor gene mask not instantiated
            self.factor_gene_mask
        except AttributeError:
            self.factor_gene_mask, self.factor_mask_keys = [], None #self._get_factor_gene_mask()

        for assay in self.schedule_assays:
            if assay == 'Direct':
                self.add_assay(
                    assays.PeakRP_Assay(
                        technology = self.isd_method, **assay_kwargs
                    )
                )
            elif assay == 'DNase' or assay == 'H3K27ac':
                self.add_assay(
                    assays.Accesibility_Assay(technology = assay, **assay_kwargs,
                        factor_gene_mask = self.factor_gene_mask,
                        selection_model = LR_BinarySearch_SampleSelectionModel(self.num_datasets_selected_anova, self.num_datasets_selected),
                        chromatin_model = LR_ChromatinModel({'C' : list(10.0**np.arange(-2,4.1,0.5))}, penalty = 'l2'), 
                        cores = self.cores, rp_map_style = self.rp_map_style
                    )
                )
            else:
                raise AssertionError('Invalid assay encountered: {}'.format(str(assay)))

    def predict(self, query_list, background_list = [], background_strategy = 'regulatory', num_background_genes = 3000, 
        seed = 2556):
        '''
    *method*
    **.predict** (self, query_list, background_list = [], background_strategy = 'regulatory', num_background_genes = 3000, seed = 2556)
    
        Predict TF influence given a set of genes.
        
        Params:
            query_list (list):
                Genes-of-interest, in either Symbol of RefSeqID format. Must provide between 20 to 500 genes.
            background_list (list):
                User-specified list of background genes to compare with query_list. Must contain more genes than query list and entire list will be used. If provided, ```background_strategy``` must be set to "provided".
            background_strategy {"regulatory","random","provided"}:
                Regulatory will sample background genes from a stratified sample of TADs and regulatory states, random will randomly sample from all non-query genes.
            num_background_genes (int):
                Number of genes to use as comparison to query genes. More background genes make test slower, but more stable.
            seed (int):
                Seed for gene selection and regression model initialization.

        Returns:
            results (lisa.core.utils.LISA_Results):
                Can be passed directly to a the pandas constructor: ``results_df = pd.DataFrame(results.to_dict())``.
            metadata (dict):
                Dictionary with test metadata. Includes query genes provided and background genes that were selected. This metadata dict also contains information on the accessibility datasets that were selected to represent the chromatin landscape around you genes-of-interest, for example, the tissue and cell line from which the profiles were derived.
        
        '''
        return super().predict(query_list, background_list=background_list, background_strategy=background_strategy, 
            num_background_genes= num_background_genes, seed=seed)



from lisa.core.data_interface import DataInterface
import argparse
import os
import numpy as np
import pandas as pd
from scipy import sparse

TECHNOLOGY = 'Motifs'

def main(species, window_size, motif_metadata, bin_sorted_hits, group_loci = 100000):

    motif_metadata = pd.read_csv(motif_metadata, sep = '\t', header = None)
    motif_metadata.columns = ['dataset_id', 'factor', 'source']
    motif_metadata = motif_metadata.set_index('dataset_id')
    motif_metadata = motif_metadata.drop_duplicates()

    data = DataInterface(species, window_size= window_size, download_if_not_exists=False,
        make_new=False, load_genes=False)

    print(data.path)
    raise Exception()

    data.create_binding_dataset(TECHNOLOGY, motif_metadata.index.values, **motif_metadata.to_dict('list'))

    id_to_idx_map = dict(zip(data.list_binding_datasets(TECHNOLOGY), np.arange(len(data.list_binding_datasets(TECHNOLOGY)))))

    current_pos = 0
    last_added_chunk = 0
    i = 0
    rows,cols,scores=[],[],[]

    with open(bin_sorted_hits, 'r') as f:

        for line in f:
            motif_id, bin_num, score = line.strip().split()

            bin_num = int(bin_num)
            
            if bin_num < current_pos:
                raise Exception('Input file not sorted!')
            elif bin_num > current_pos and i >= group_loci:
                print('Adding matrix segment ...')
                matrix_form = sparse.coo_matrix((scores, (rows, cols))).tocsr()
                data.append_csr(TECHNOLOGY, matrix_form)
                last_added_chunk = bin_num
                i=0
                rows,cols,scores=[],[],[]

            tf_idx = id_to_idx_map[motif_id]
            rows.append(bin_num - last_added_chunk)
            cols.append(tf_idx)
            scores.append(int(score))
            current_pos = bin_num
            i+=1
        
        if len(rows) > 0:
            matrix_form = sparse.coo_matrix((scores, (rows, cols))).tocsr()
            data.append_csr(TECHNOLOGY, matrix_form)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Saves indices lists to factor binding h5. Filename specifies metadata: \{technology\}-\{dataset_id\}-\{metadata1_metadata2_...\}')
    parser.add_argument('species', type = str, choices = ['hg38','mm10'])
    parser.add_argument('window_size', type = int)
    parser.add_argument('motif_metadata', type = str)
    parser.add_argument('hits', type = str)

    args = parser.parse_args()

    main(args.species, int(args.window_size), args.motif_metadata, args.hits)



from lisa.core.data_interface import DataInterface
import argparse
import os
import numpy as np
import pandas as pd

def main(species, window_size, cistrome_metadata, motif_metadata, index_files):

    cistrome_metadata = pd.read_csv(cistrome_metadata, sep = '\t').set_index('DCid')
    cistrome_metadata.index = cistrome_metadata.index.astype(str)
    motif_metadata = pd.read_csv(motif_metadata, sep = '\t', header = None)
    motif_metadata.columns = ['dataset_id', 'factor', 'source']
    motif_metadata = motif_metadata.set_index('dataset_id')
    motif_metadata = motif_metadata.drop_duplicates()

    data = DataInterface(species, window_size= window_size, download_if_not_exists=False,
        make_new=False, load_genes=False)

    dataset_ids = motif_metadata.index.values

    data.create_binding_dataset('Motif', dataset_ids)

    

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Saves indices lists to factor binding h5. Filename specifies metadata: \{technology\}-\{dataset_id\}-\{metadata1_metadata2_...\}')
    parser.add_argument('species', type = str, choices = ['hg38','mm10'])
    parser.add_argument('window_size', type = int)
    parser.add_argument('cistrome_metadata', type = str)
    parser.add_argument('motif_metadata', type = str)
    parser.add_argument('index_files', type = str, nargs='+')

    args = parser.parse_args()

    main(args.species, int(args.window_size), args.cistrome_metadata, args.motif_metadata, args.index_files)


from urllib import request
import gzip
from lisa.core.data_interface import DataInterface
from lisa.core.genome_tools import Region, BadRegionError
from lisa.core.utils import Log, LoadingBar
import argparse
from scipy import sparse
import os
from scipy.stats import gamma
import numpy as np
from sys import stdout, stderr

def main(*,species, motif_bed, window_size, dataset_id, output):

    genome = DataInterface.load_genome(species, window_size)

    factor_name = None
    window_nums, scores = [],[]

    #adjust p-val cutoff based on the filesize (only affects p-val if file is huge)
    pval_cutoff = 430

    with open(output, 'w') as o:

        with gzip.open(motif_bed, 'rb') as bed:

            for i, line in enumerate(bed):
                
                chrom, start, end, factor, relscore, log_pval, strand = line.decode('utf-8').strip().split('\t')
                
                if i == 0:
                    factor_name = factor
                    print('Binning {} motifs with pval cutoff of {} ...'.format(factor_name.upper(), str(pval_cutoff)), file = stderr)

                neg_log10_pval = int(log_pval)

                if neg_log10_pval >= pval_cutoff:

                    try:
                        hit_windows = genome.get_region_windows(Region(chrom, start, end))
                        
                        for hit_window in hit_windows:
                            print(dataset_id, hit_window, neg_log10_pval, sep = '\t', file = o)                    

                    except BadRegionError:
                        pass
    
    print(dataset_id, factor_name.upper(), 'JASPAR', sep = '\t')

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument('species', type = str, choices = ['hg38','mm10'])
    parser.add_argument('motif_bed', type = str)
    parser.add_argument('-w', '--window_size', type = int, required=True)
    parser.add_argument('-o', '--output', type = str, required=True)
    parser.add_argument('-i','--id', type = str, required=True)
    args = parser.parse_args()

    main(species = args.species, motif_bed = args.motif_bed, 
        window_size = args.window_size, dataset_id = args.id, output = args.output)



#lisa modules
import lisa.core.gene_selection as gene_selection
from lisa.core.utils import LoadingBar, Log, LISA_Results
from lisa._version import __version__
from lisa._version import __file__ as version_file
from lisa.core.data_interface import DataInterface, PACKAGE_PATH, DatasetNotFoundError

#standard library
import configparser
from collections.abc import Iterable
import sys
import json
import multiprocessing
from urllib import request, error
import tarfile
import os
import sys

#required
import numpy as np
import h5py as h5
from scipy import sparse, stats
#from pyranges.pyranges import PyRange

CONFIG_PATH = os.path.join(os.path.dirname(__file__), 'config.ini')

class DownloadRequiredError(BaseException):
    pass

class LISA_Core:

    @classmethod
    def get_dataset_url(cls, species):
        return DataInterface.get_dataset_url(species, cls.window_size)

    @classmethod
    def download_dataset(cls, species):
        DataInterface.fetch_from_cistrome(species, cls.window_size)

    @classmethod
    def get_dataset_path(cls, species):
        return DataInterface.get_dataset_path(species, cls.window_size)

    #class dictionary, converts cmd line alias to standard symbols for knockout methods
    factor_binding_technologies = dict(
        chipseq = 'ChIP-seq',
        motifs = 'Motifs'
    )

    def __init__(self, species, config, window_size,
        isd_method = 'chipseq',
        verbose = True,
        log = None,
    ):
        
        self._config = config
        #all paramter checking is done on LISA instantiation to ensure consistency between python module and cmd line usage
        self.isd_options = self._config.get('lisa_params', 'isd_methods').split(',')
        self.background_options = self._config.get('lisa_params', 'background_strategies').split(',')
        self.max_query_genes = int(self._config.get('lisa_params', 'max_user_genelist_len'))
        assert( isd_method in  self.isd_options ), 'ISD method must be ({})'.format(', '.join(self.isd_options))
        assert( species in ['mm10','hg38'] ), "Species must be either hg38 or mm10"
        
        self.isd_method = self.factor_binding_technologies[isd_method]

        #use provided log object or create a new one. Applications which call LISA may want to pass in their own log object for control of indentation
        if log is None:
            self.log = Log(sys.stderr, verbose = verbose)
        else:
            assert( isinstance(log, Log) )
            self.log = log

        self.data_interface = DataInterface(species, window_size=window_size, make_new=False,
            download_if_not_exists=True, log=self.log, load_genes=True)
        
        #data has not yet been loaded
        self._is_loaded = False
        self.assays = []

    def _load_rp_map(self):
        raise NotImplementedError()

    def _load_factor_binding_data(self):
        raise NotImplementedError()

    def _load_data(self):

        if self._is_loaded:
            raise AssertionError('Data is already loaded')

        with self.log.section('Loading data into memory (only on first prediction):') as log:

            log.append('Loading regulatory potential map ...')
            self.rp_map = self._load_rp_map()

            with log.section('Loading binding data ...') as log:
                self.factor_binding, self.factor_dataset_ids, self.factor_metadata = self._load_factor_binding_data()

            #self.factor_metadata = self.link_metadata(self.isd_method).select(self.factor_dataset_ids)
            self.log.append('Done!')

        self._is_loaded = True

    #____ GENE SELECTION FUNCTIONS ____
    def _preprocess_gene_list(self, genes):
        #make sure gene lists are composed of stripped strings
        return [str(gene).strip() for gene in genes if len(gene) > 0]
        
    def _sample_background_genes(self, query_genes, background_strategy = 'regulatory', num_background_genes = 3000, seed = None):

        assert( (background_strategy == 'all' or num_background_genes >= len(query_genes)) and len(query_genes) <= 19000//2 ), "More query genes selected than background genes"

        background_candidates = self.data_interface.genes.get_distinct_genes_by_symbol(excluding = query_genes.get_symbols())

        if background_strategy == 'all':
            return background_candidates
        else:
            assert(len(background_candidates) >= num_background_genes), 'Number of background candidates must be greater than or equal number of genes to select as background.'
            #if no down-sampling is needed:
            if len(background_candidates) == num_background_genes:
                background_genes = background_candidates

            if background_strategy == 'regulatory':

                background_genes = background_candidates.sample_by_TAD(num_background_genes, seed = seed)

                if len(background_genes) > num_background_genes:
                    background_genes = background_genes.random_sample(num_background_genes, seed = seed)

            elif background_strategy == 'random':
                background_genes = background_candidates.random_sample(num_background_genes, seed = seed)
            else:
                raise AssertionError('Background selection strategy {} not supported'.format(background_strategy))

        return background_genes

    def _get_query_and_background_genes(self, query_list, *, background_list = [], num_background_genes = 3000, background_strategy =
        'regulatory', seed = None):

        query_list = self._preprocess_gene_list(query_list)
        background_list = self._preprocess_gene_list(background_list)

        query_genes = self.data_interface.genes.match_user_provided_genes(query_list)

        assert(20 <= len(query_genes) <= self.max_query_genes), 'User must provide list of 20 to {} unique genes. Provided {}'\
            .format(str(self.max_query_genes), str(len(query_genes)))

        if background_strategy == 'provided':
            
            background_matches = self.data_interface.genes.match_user_provided_genes(background_list)
            background_genes = background_matches.get_distinct_genes_by_symbol(excluding = query_genes.get_symbols())
            assert( len(background_genes) > len(query_genes) ), 'Number of background genes must exceed number of query genes provided'
        
        else:
            background_genes = self._sample_background_genes(query_genes, background_strategy = background_strategy, 
                num_background_genes = num_background_genes, seed = seed)

        return query_genes, background_genes

    def _make_gene_mask(self, query_genes, background_genes):

        label_dict = { 
            gene.get_location() : (int(i < len(query_genes)), gene.get_name())
            for i, gene in enumerate(list(query_genes) + list(background_genes))
        }

        gene_mask, query_names, background_names, label_vector = [],[],[],[]
        for loc in self.data_interface.rp_map_locs:
            gene_mask.append(loc in label_dict)
            
            if loc in label_dict:
                label_vector.append(label_dict[loc][0])

                if label_dict[loc][0] == 0:
                    background_names.append(label_dict[loc][1])
                else:
                    query_names.append(label_dict[loc][1])


        return np.array(gene_mask), np.array(label_vector).astype(bool), dict(
            query_symbols = query_names,
            background_symbols = background_names,
        )
        
    #KEYSTONE FUNCTION FOR GENE SELECTION
    def _choose_genes(self, query_list, background_list = [], background_strategy = 'regulatory', num_background_genes = 3000, seed = None):
        #validate user args
        if background_list is None:
            background_list = []
        assert( isinstance(num_background_genes, int) )
        assert( background_strategy in  self.background_options ), 'Background strategy must be in \{{}\}'.format(', '.join(self.background_options))
        assert( isinstance(query_list, Iterable) )
        assert( isinstance(background_list, Iterable))
        #check to make sure data is downloaded

        with self.log.section('Matching genes and selecting background ...'):

            if len(background_list) > 0 and background_strategy == 'provided':
                self.log.append('User provided background genes!')

            query_genes, background_genes = self._get_query_and_background_genes(query_list, background_list = background_list, 
                    num_background_genes = num_background_genes, background_strategy = background_strategy, seed = seed)
            
            gene_mask, label_vector, gene_info_dict = self._make_gene_mask(query_genes, background_genes)

        return gene_mask, label_vector, gene_info_dict

    #____ ASSAY INSTANTIATION AND CALLING
    def add_assay(self, assay):
        self.assays.append(assay)

    def _run_assays(self, gene_mask, label_vector, *args, **kwargs):
        #run each assay specified in the _load_data step. Each assay returns a p_value for each factor binding sample, and returns metadata
        #each assay takes the same arguments and makes the same returns but can perform different tests
        assay_pvals, assay_info = {},{}
        for assay in self.assays:
            assay_pvals[assay.technology + '_p_value'], assay_info[assay.technology] = assay.predict(gene_mask, label_vector, *args, **kwargs)

        return assay_pvals, assay_info

    def _initialize_assays(self):
        raise NotImplementedError('_initialize_assays function is not implemented on base LISA object') 

    def _initialize_resources(self):
        try:
            self._check_for_data()
        except DownloadRequiredError:
            self.download_data()
            self._check_for_data()

        #self.log.append('Using {} cores ...'.format(str(self.cores)))

        if not self._is_loaded:
            self._load_data()

    #___ RESULTS SUMMARIZATION AND FORMATTING____
    @staticmethod
    def _combine_tests(p_vals, weights=None):
        #https://arxiv.org/abs/1808.09011
        #combine p-values from all assays
        p_vals = np.array(p_vals).astype(np.float64)

        assert(len(p_vals.shape) == 2 ), 'P-values must be provided as matrix of (samples, multiple p-values)'
        assert(p_vals.shape[1] > 1), 'Must have multiple p-values to combine'

        if weights is None:
            weights = 1 / p_vals.shape[1]
        else:
            assert(p_vals.shape[1] == weights.shape[1])

        individual_statistics = np.where(p_vals > 1e-15, np.tan((0.5-p_vals) * np.pi), 1/p_vals/np.pi)
        test_statistic = np.sum(weights * individual_statistics, axis = 1)
        
        # lim x->inf of arctan(x) = pi/2
        # so for large x, arctan(x) ~ pi/2 - 1/x, as 1/x -> 0
        # substitute this arctan approx when x > 1e15

        too_large = test_statistic > 1e15
        combined_p_value = np.where(~too_large, 0.5 - np.arctan(test_statistic)/np.pi, 1/test_statistic/np.pi)

        return test_statistic, combined_p_value

    def _format_results(self, assay_pvals, assay_info, gene_info_dict, **kwargs):
        self.log.append('Formatting output ...')
        #instantiate a lisa results object
        results = LISA_Results.fromdict(
            **self.factor_metadata,
            **assay_pvals,
            **kwargs
        )

        results = results.sortby('summary_p_value', add_rank = True, reverse = False)
        
        self.log.append('Done!')
        
        #return model_metadata as big dictionary
        return results, dict(
            **gene_info_dict,
            **assay_info,
        )

    def _summarize_p_values(self, assay_pvals):
        
        if len(assay_pvals) > 1:
            #combine p-values from all assays on a per-sample basis
            with self.log.section('Mixing effects using Cauchy combination ...'):

                aggregate_pvals = np.array(list(assay_pvals.values())).T
                
                combined_statistic, summary_p_value = self._combine_tests(aggregate_pvals)
        else:
            summary_p_value = np.array(list(assay_pvals.values())).reshape(-1)

        return summary_p_value


    #___ CLASS PREDICTION FUNCTION (USER INTERFACE) ___
    def predict(self, query_list, *args, background_list = [], background_strategy = 'regulatory', num_background_genes = 3000, seed = 2556, **kwargs):
        
        try:

            gene_mask, label_vector, gene_info_dict = self._choose_genes(query_list, background_list, background_strategy, num_background_genes, seed)

            self.log.append('{} query genes and {} background genes selected'\
                .format(str(label_vector.sum()), str(len(label_vector) - label_vector.sum())))

            if not self._is_loaded:
                self._load_data()

            if len(self.assays) == 0:
                #the initialize assay section is the mutable section for each interface
                self._initialize_assays(data_interface = self.data_interface, log = self.log,
                    rp_map = self.rp_map, factor_binding = self.factor_binding, 
                    factor_dataset_ids = self.factor_dataset_ids)
            
            assay_pvals, assay_info = self._run_assays(gene_mask, label_vector, *args, **kwargs)

            summary_p_value = self._summarize_p_values(assay_pvals)

            return self._format_results(assay_pvals, assay_info, gene_info_dict, summary_p_value = summary_p_value)

        except (DatasetNotFoundError):
            raise DownloadRequiredError('Data is malformed or incomplete, run "lisa download [species] [test]" to redownload dataset')

    def get_docs(self):
        return '\n'.join([
            self.__doc__,
            self.__init__.__doc__,
            self.predict.__doc__]
        )



import pandas as pd
from lisa import FromGenes
import requests
import fire
import argparse
import numpy as np

URL = 'http://lisa.cistrome.org/gallery/{id}_{dir}.txt'

TEST = [
    ('189', ['NFE2']),
    ('172',['PPARG']),
    ('187',['YY1']),
    ('220',['RUNX3']),
    ('135', ['HNF4A']),
    ('154',['MYOD1']),
    ('173',['EBF1']),
    ('182',['NR1D1']),
    ('263',['ESR1']),
    ('264',['GATA3']),
    ('183',['NKX2-5']),
    ('163',['GATA4']),
    ('280',['MED1']),
    ('37',['CEBPG']),
    ('304',['RUNX1']),
    ('275', ['BMI1']),
    ('222', ['BCL6']),
    ('292',['SOX2']),
    ('20',['IRF4']),
    ('181', ['FOXP3'])
]

def get_ranks(factor_name, results, rank_column = 'summary_p_value', max_rank = 500):

    results = pd.DataFrame(results.to_dict())

    results['col_rank'] = results[rank_column].rank()

    ranks = np.array(results[results.factor == factor_name].col_rank.values)

    ranks = np.clip(ranks, None, max_rank + 1)

    return list(ranks)

def get_genelist(cistrome_id, direction):

    r = requests.get(URL.format(id = cistrome_id, dir = direction))

    return [x.strip() for x in r.text.split('\n')]


def run_test(lisa_obj, factor_name, cistrome_id, direction, max_rank = 500, rank_column = 'summary_p_value'):
    
    genelist = get_genelist(cistrome_id, direction)

    try:

        results, _ = lisa_obj.predict(genelist, num_background_genes = 750, seed=2556)

        ranks = get_ranks(factor_name, results, rank_column=rank_column, max_rank=max_rank)

        print(factor_name + ': ', str(ranks))

        return ranks
    except Exception as err:
        print(repr(err))
        return []

def run_tests(lisa_obj, tests, max_rank = 500, rank_column = 'summary_p_value'):

    all_ranks = []
    for test in tests:

        for direction in ['up','down']:
            ranks = run_test(lisa_obj, test[1][0], test[0], direction, max_rank= max_rank, rank_column=rank_column)
            all_ranks.extend(ranks)

    return all_ranks


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument('output', type = argparse.FileType('w'))
    args = parser.parse_args()

    lisa_obj = FromGenes('mm10', assays = ['DNase'], verbose=1)

    all_ranks = run_tests(lisa_obj, TEST, max_rank=500, rank_column='DNase_p_value')

    print('\n'.join([str(r) for r in all_ranks]), end = '', file = args.output)



import numpy as np
import os
import random
from collections import Counter, defaultdict, OrderedDict
import random
from lisa.core.genome_tools import Region


class Gene:
    '''
    A gene has a unique genomic region, an accepted name, and aliases that also correspond to that gene or genomic region
    '''
    def __init__(self, chrom, tss_start, tss_end, names, tad_domain = None):
        self.chrom = chrom
        self.start = int(tss_start)
        self.end = int(tss_end)
        self.aliases = []
        self.tad_domain = tad_domain
        if isinstance(names, str):
            self.add_alias(names.upper())
        else:
            self.add_aliases([name.upper() for name in names])
        self.location = ':'.join([str(property_) for property_ in (self.chrom, self.start, self.end)])
        
    def add_alias(self, alias, is_name = True):
                
        if not alias in self.aliases:
            self.aliases.append(alias)

    def get_name(self):
        return self.aliases[0]

    def add_aliases(self, aliases):
        assert( isinstance(aliases, list) )
        for alias in aliases:
            self.add_alias(alias)

    def get_location(self):
        return self.location

    def get_tss_region(self):
        return Region(self.chrom, self.start, self.end, annotation = self)
    
    def __eq__(self, other):
        if isinstance(other, str):
            return other in self.aliases
        elif isinstance(other, Gene):
            return self.get_location() == other.get_location()
        else:
            return False

    def __repr__(self):
        return '\t'.join([str(x) for x in [self.get_location(), self.aliases[0], self.tad_domain, '|'.join(self.aliases)]])

    def __str__(self):
        return self.get_name()

    def get_RP_signature(self, bins, bin_index, delta = 10000, max_influence_distance = 100000):

        tss = self.start
        
        #find bin regions defining interesting RP region
        min_bin, split_bin, max_bin = np.digitize([tss - max_influence_distance, tss, tss + max_influence_distance], bins)
        #subset interesting region of chrom
        bin_indices = bin_index[min_bin: max_bin - 1]
        #split the bin holding the TSS into two bins
        bins = np.concatenate([bins[min_bin: split_bin], [tss], bins[split_bin: max_bin]])
        split_bin -= min_bin
        tss_bins = (split_bin - 1, split_bin)

        #get bin intervals to the left and right of TSS, then concatenate
        left_bins = np.abs(np.array(list(zip(bins[1:tss_bins[1] + 1], bins[:tss_bins[0] + 1]))) - tss)
        right_bins = np.abs(np.array(list(zip(bins[tss_bins[1]:-1], bins[tss_bins[1] + 1:]))) - tss)
        intervals = np.concatenate([left_bins, right_bins], axis = 0)

        #get integral of RP at boundary locations
        RP = intervals * (-np.log(1/3) / delta)
        RP = 2 * ( RP - np.log(np.exp(RP) + 1))
        #compute RP over area
        RP = np.subtract(RP[:,1], RP[:,0])
        #sum bins split by TSS
        summed_tss_rp = RP[list(tss_bins)].sum()
        RP[tss_bins[0]] = summed_tss_rp
        #remove split bin
        RP = np.delete(RP, tss_bins[1])
        #normalize so bin with TSS has RP of 1
        RP = RP / RP.max()

        return RP, bin_indices


class RefSeqGene(Gene):

    promoter_width_from_tss = 1500

    def __init__(self, name, chrom, strand,
        txStart, txEnd, exonStarts, exonEnds, symbol, tad_cluster, *, genome):

        if strand == '-':
            tss = (int(txEnd), int(txEnd) + 1)
        else:
            tss = (int(txStart), int(txStart) + 1)
            
        super().__init__(chrom, *tss, [symbol.upper(), name.upper()], tad_domain=tad_cluster)

        self.special_regions = dict()

        self.add_region(Region(chrom, *tss).slop(self.promoter_width_from_tss, genome))
        if exonStarts != '' and exonEnds != '':
            for exon_start, exon_end in zip(exonStarts.strip(',').split(','), exonEnds.strip(',').split(',')):
                self.add_region(Region(chrom, exon_start, exon_end))

        self.strand = strand
        self.is_noncoding = name[:3] == "NR_"

    def get_exon_regions(self):
        return list(self.special_regions.values())

    def add_region(self, region):
        if not region.to_tuple() in self.special_regions:
            self.special_regions[region.to_tuple()] = region

    def add_regions(self, new_regions):
        for new_region in new_regions:
            self.add_region(new_region)


class GeneSet:
    '''
    Enforces gene organization rules: 
        A genomic location may correspond to many names
        A name may correspond to many locations
        Primary organization should be by genomic location
    '''

    @classmethod
    def from_file(cls, path):

        new_geneset = cls()

        with open(path, 'r') as f:
            for line in f.readlines()[1:]:
                location, name, tad, aliases = [x.strip() for x in line.split('\t')]
                new_gene = Gene(*location.split(':'), aliases.split('|'), tad_domain = tad)
                new_geneset.add_gene(new_gene)

        return new_geneset

    @classmethod
    def from_refseq(cls, path, genome):
        
        new_geneset = cls()
        with open(path, 'r') as f:
            for line in f.readlines():
                #print(line.strip().split('\t'))
                new_geneset.add_gene(RefSeqGene(*[x.strip() for x in line.strip().split('\t')], genome = genome))

        return new_geneset
            
    def __init__(self):
        self.genes_by_name = defaultdict(list)
        self.genes_by_chr = OrderedDict()

    def add_genes(self, new_genes):
        for new_gene in new_genes:
            self.add_gene(new_gene)

        return self

    def add_gene(self, new_gene):
        
        #finds if gene location is already inhabited
        if new_gene.get_location() in self.genes_by_chr:
            #get existing gene object
            existing_gene = self.genes_by_chr[new_gene.get_location()]
            #adds the new names to the existing object for this genomic location / gene
            existing_gene.add_aliases(new_gene.aliases)

            try:
                existing_gene.add_regions(new_gene.special_regions)
            except AttributeError:
                pass

            #adds pointers from these names to the existing gene
            for alias in new_gene.aliases:
                #if this alias is not registered
                if not alias in self.genes_by_name:
                    #add the location under the alias
                    self.genes_by_name[alias.upper()].append(existing_gene)

        else:
            #add this gene under its genomic location
            self.genes_by_chr[new_gene.get_location()] = new_gene

            #for its names, add this location
            for alias in new_gene.aliases:
                #uppercase the gene name so that capitalization is not a factor 
                self.genes_by_name[alias.upper()].append(new_gene)        

        return self

    def get_symbols(self):
        return [gene.get_name() for gene in self]

    def get_locations(self):
        return [gene.get_location() for gene in self]

    def get_distinct_genes_by_symbol(self, excluding = set(), exclude_nc_rna = True):

        names = self.get_symbols()

        distinct_names = set(names).difference(excluding)

        distinct_genes = GeneSet()
        for dinstinct_name in distinct_names:
            add_gene = self.get_gene_by_name(dinstinct_name)
            try:
                if not add_gene.is_noncoding:
                    distinct_genes.add_gene(add_gene)
            except AttributeError:
                distinct_genes.add_gene(add_gene)

        return distinct_genes


    def get_gene_by_name(self, name):

        name = name.upper()
        if name not in self.genes_by_name:
            raise KeyError()
        else:
            return self.genes_by_name[name][0]


    def __str__(self):
        return '\t'.join(['location','gene_name','tad_domain','aliases']) + '\n' + '\n'.join([repr(gene) for gene in self.genes_by_chr.values()])


    def get_genes_by_chrom(self, chromosome):
        return [
            gene for location_key, gene in self.genes_by_chr.items() if location_key.split(':')[0] == chromosome
        ]

    def __len__(self):
        return len(self.genes_by_chr)

    def __iter__(self):
        return iter(list(self.genes_by_chr.values()))


    def match_user_provided_genes(self, user_genelist):
        rejects = []
        selected_genes = GeneSet()
        for gene_candidate in user_genelist:
            try:
                selected_genes.add_gene( self.get_gene_by_name(gene_candidate) )
            except KeyError:
                rejects.append(gene_candidate)
        
        return selected_genes.get_distinct_genes_by_symbol()


    def random_sample(self, sample_num, seed = None):
        if not seed is None:
            np.random.seed(seed)
        
        assert(len(self) > sample_num), 'Background gene list provided must contain more than {} genes'.format(str(sample_num))
        
        #if same number of genes, skip sample
        if len(self) == sample_num:
            return self
        
        else:
            return GeneSet().add_genes(np.random.choice(sorted(list(self), key = lambda x : x.location), sample_num, replace = False))

    #enforce sampling according to the TAD distribution of the genome.
    def sample_by_TAD(self, sample_num, seed = None):
        
        if not seed is None:
            np.random.seed(seed)

        #intersect background gene list with TAD_list to eliminate reserved genes, sorting to maintain order for seed repeatability
        TAD_data = [(gene, gene.tad_domain) for gene in sorted(self, key = lambda x : x.location)]

        #collect list of genes in each bin
        genes_in_TAD = defaultdict(list)
        for gene, tad_group in TAD_data:
            genes_in_TAD[tad_group].append(gene)
        
        #calculates number of genes in each TAD group
        num_genes_in_TAD = {tad_group : len(genes) for tad_group, genes in genes_in_TAD.items()}

        #calculates the number of genes expected to be sampled from each TAD, with a bit of an enrichment (1.1x). Ensure all TADs have >= 1 expected genes
        expected_samples = {
            tad_group : max(1, int(num_genes / len(TAD_data) * sample_num * 1.1))
            for tad_group, num_genes in num_genes_in_TAD.items()
        }

        #samples the expected number of genes from the TAD (or the number of genes if expected is > actual)
        selected_genes = GeneSet()
        for tad_group, num_expected in expected_samples.items():
            sampled_genes = np.random.choice(sorted(genes_in_TAD[tad_group], key = lambda x : x.location), min(num_genes_in_TAD[tad_group] - 1, num_expected), replace = False)
            selected_genes.add_genes(sampled_genes)

        return selected_genes



from numpy import inf

def parse_deseq_file(path, lfc_cutoff = 1.5, pval_cutoff = 0.05, sep = '\t'):

    up_genes, down_genes = [],[]

    with open(path, 'r') as f:
        
        skipped_header = False
        for i, line in enumerate(f):

            if i > 0 and not (line == '' or line == '\n'):
                fields = line.strip().split('\t')
                assert(len(fields) == 7), 'Correctly formatted DESeq2 file must have 7 fields.'

                try:
                    gene_name, lfc, pval = fields[0], float(fields[2]), float(fields[6])

                    if pval <= pval_cutoff and abs(lfc) >= lfc_cutoff:
                        if lfc > 0:
                            up_genes.append(gene_name)
                        elif lfc < 0:
                            down_genes.append(gene_name)

                except ValueError:
                    pass

    return up_genes, down_genes


def parse_macs_file(path):
    '''
**lisa.parse_macs_file** (path)

Parse regions and region scores from MACS2 .xls output file. Region scores are the number of reads at an associated position (pileup field)

Params:
    path (str):
        path to MACS2 .xls output file
    
Returns:
    region_fields (list):
        list of tuples of (chr, start, end) parsed from MACS file.
    region_scores (list):
        list of scores, read depth at each position

    '''

    with open(path, 'r') as bed:
        lines = bed.readlines()

    region_fields = []
    region_scores = []
    found_header = False
    for line in lines:
        if found_header:
            line = line.split('\t')
            assert(len(line) == 10), 'Correctly-formatted MACS2 .xls files have 10 fields.'
            region_fields.append(line[:3])
            region_scores.append(line[5])
        elif not line[0] == "#" and line[:3] == 'chr':
            line = line.split('\t')
            assert(len(line) == 10), 'Correctly-formatted MACS2 .xls files have 10 fields.'
            found_header = True

    return region_fields, region_scores


def parse_regions_file(path, header = False):
    '''
**lisa.parse_regions_file** (path, header = False)

Parse regions and region scores from bed-like file. Must have columns *chrom, start, end* with optional fourth *score* column.

Params:
    path (str):
        path to MACS2 .xls output file
    header (bool):
        If true, skip first line of bedfile
    
Returns:
    region_fields (list):
        list of tuples of (chr, start, end) parsed from MACS file.
    region_scores (list):
        list of scores, read depth at each position
    '''
    with open(path, 'r') as bed:
        lines = bed.readlines()

    assert(type(header) == bool)

    region_fields, region_scores = [],[]
    num_fields = len(lines[0].split('\t'))
    assert(num_fields in [3,4]), 'Bedfile must have only three columns (chr,start,end) or four (chr,start,end,score)'

    for i, line in enumerate(lines[(1 if header else 0) : ]):
        
        line = line.strip().split('\t')
        if len(line) != num_fields:
            raise AssertionError('Error at line #{}: expected {} fields, got {}'\
                .format(str(i), str(num_fields), str(len(line))))
        
        region_fields.append(line[:3])
        if num_fields == 4:
            region_scores.append(line[-1])
        elif num_fields == 3:
            region_scores.append(1)

    return region_fields, region_scores


def parse_bedfile(path, header = False):
    
    regions = []

    with open(path, 'r') as bed:
        for line in bed.readlines()[(1 if header else 0) : ]:
            fields = line.strip().split('\t')
            assert(len(fields) >= 3)
            chrom, start, end = fields[:3]
            regions.append((chrom, start, end))

    return regions



import argparse
from lisa.core.data_interface import DataInterface
import numpy as np
import pandas as pd
import os

def main(args):

    cistrome_metadata = pd.read_csv(args.cistrome_metadata, sep = '\t').set_index('DCid')
    cistrome_metadata.index = cistrome_metadata.index.astype(str)

    data = DataInterface(args.species, window_size= args.window_size, download_if_not_exists=False,
        make_new=False, load_genes=True)

    rp_map_styles = data.get_rp_maps()

    if len(rp_map_styles) == 0:

        basic_rp_map, enhanced_rp_map = data.build_binned_rp_map('basic',10000), data.build_binned_rp_map('enhanced', 10000)

        data.add_rp_map('basic_10K', basic_rp_map)
        data.add_rp_map('enhanced_10K', enhanced_rp_map)

    else:

        basic_rp_map = data.get_rp_map('basic_10K')
        enhanced_rp_map = data.get_rp_map('enhanced_10K')


    for arr_name in args.coverage_arrays:
        
        coverage_array = np.load(arr_name)
        
        technology, dataset_id = os.path.basename(arr_name).split('_')

        dataset_id = '.'.join(dataset_id.split('.')[:-1])

        metadata_headers = data.get_metadata_headers(technology)

        meta_dict = cistrome_metadata.loc[dataset_id, metadata_headers].to_dict()

        data.add_profile_data(technology, dataset_id, coverage_array, [basic_rp_map, enhanced_rp_map],
            ['basic_10K','enhanced_10K'], **meta_dict)


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument('species', type = str, choices = ['hg38','mm10'])
    parser.add_argument('window_size', type = int)
    parser.add_argument('cistrome_metadata', type = str)
    parser.add_argument('coverage_arrays', type = str, nargs = '+')

    args = parser.parse_args()

    main(args)



from lisa.core.data_interface import DataInterface
import argparse
import os
import numpy as np
import pandas as pd

def main(species, window_size, cistrome_metadata, motif_metadata, index_files):

    cistrome_metadata = pd.read_csv(cistrome_metadata, sep = '\t').set_index('DCid')
    cistrome_metadata.index = cistrome_metadata.index.astype(str)
    motif_metadata = pd.read_csv(motif_metadata, sep = '\t', header = None)
    motif_metadata.columns = ['dataset_id', 'factor', 'source']
    motif_metadata = motif_metadata.set_index('dataset_id')
    motif_metadata = motif_metadata.drop_duplicates()

    data = DataInterface(species, window_size= window_size, download_if_not_exists=False,
        make_new=False, load_genes=False)

    for index_file in index_files:
        
        with open(index_file, 'r') as f:
            hit_bins = np.array([int(ind.strip()) for ind in f.readlines()])
        
        technology, dataset_id = os.path.basename(index_file).split('_')

        dataset_id = '.'.join(dataset_id.split('.')[:-1])

        metadata_headers = data.get_metadata_headers(technology)

        if technology == 'Motifs':
            meta_dict = motif_metadata.loc[dataset_id, metadata_headers].to_dict()
            meta_dict['source'] = 'jaspar'
        else:
            meta_dict = cistrome_metadata.loc[dataset_id, metadata_headers].to_dict()

        data.add_binding_data(technology, dataset_id, hit_bins, **meta_dict)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Saves indices lists to factor binding h5. Filename specifies metadata: \{technology\}-\{dataset_id\}-\{metadata1_metadata2_...\}')
    parser.add_argument('species', type = str, choices = ['hg38','mm10'])
    parser.add_argument('window_size', type = int)
    parser.add_argument('cistrome_metadata', type = str)
    parser.add_argument('motif_metadata', type = str)
    parser.add_argument('index_files', type = str, nargs='+')

    args = parser.parse_args()

    main(args.species, int(args.window_size), args.cistrome_metadata, args.motif_metadata, args.index_files)